SystemOrganization addCategory: #'Prolog-Compiler'!


Decompiler subclass: #LDecompiler
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Prolog-Compiler'!

LDecompiler class
	instanceVariableNames: ''!

Parser subclass: #LParser
	instanceVariableNames: 'prologMethodSelector prologMethodArgument cuts temporaries statements'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Prolog-Compiler'!

Compiler subclass: #LCompiler
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Prolog-Compiler'!

LCompiler class
	instanceVariableNames: ''!

LParser class
	instanceVariableNames: ''!


!LCompiler commentStamp: 'spfa 6/20/2008 12:39' prior: 0!
Compiler for Prolog class tree!
]style[(30)!

!LDecompiler commentStamp: 'spfa 6/20/2008 12:41' prior: 0!
Decompiler for Prolog class tree

do not produce actual working code for prolog methods because of the selector dispatching system. see usages of #selectorForPredicate:!
]style[(168)!

!LParser commentStamp: 'bolot 12/13/2002 15:14' prior: 0!
RE: safeGuardReturnBlock
Makes sure that the method returns, even if no match occurs.

RE: queryUndefined
Disabled for Prolog methods (because undefined Prolog variables are unbounds).

RE: prologMethodSelector
Can be used to determine if this is a smalltalk method or prolog method.

RE: Smalltalk guys
Strings, symbols, numbers, and literal points (45@65) are recognized.

RE: evaluation
Support for nicer, more Prolog-like syntax for evaluation is done through direct use of LCompiler.
Example: LCompiler evaluate: 'Family new :? father(''John'', x).'
(see Family comment for current situation)

This class and other related classes in this category were originally written by or are based on the work by Mike Teng <miket@objectshare.com>, then at Digitalk, as part of the Prolog/V implementation distributed with Smalltalk V/286 in 1988.
We gratefully acknowledge Mike's permission to port his code to Squeak.

The port to Squeak is the work of Bolot Kerimbaev <bolot@cc.gatech.edu>, a summer intern at Walt Disney Imagineering.!
]style[(1032)!

!LParser methodsFor: 'prolog parser' stamp: 'spfa 11/18/2009 18:10'!
addTemp: aTerm
	"Decide whether aTerm should be declared as a temporay."
	| var |

	var := encoder encodeVariable: aTerm 
		ifUnknown: [encoder bindTemp: aTerm].

	(#('true' 'false' 'nil' 'super' 'self') includes: aTerm) ifTrue: [^var].
	aTerm first isLowercase ifFalse: [^var].
	(temporaries includes: aTerm) ifTrue: [^var].

	temporaries add: var.
	statements addFirst: 
		(AssignmentNode new
			variable: var
			value: (self send: #new	to: 'LogicRef')).
	^var! !

!LParser methodsFor: 'prolog parser' stamp: 'spfa 11/18/2009 18:10'!
argumentList: aList
	^ aList isConsCell
		ifTrue: [self arguments: aList]
		ifFalse: [aList]! !

!LParser methodsFor: 'prolog parser' stamp: 'spfa 11/18/2009 18:10'!
arguments: aList
	"Answer a MessageNode for predicate arguments."
	| receiver bList |

	aList ifNil: [^ nil] "zero argument".

	receiver := self send: #car:
			to: (aList isRelation ifTrue: ['Relation'] ifFalse: ['PrologList'])
			with: {self argumentList: aList head}.

	aList cdr ifNil: [^ receiver] "single argument". 

	aList cdr isConsCell ifFalse: ["[x | y]"
		^ self send: #car:cdr:
			to: (aList isRelation ifTrue: ['Relation'] ifFalse: ['PrologList'])
			with: {self argumentList: aList head . self argumentList: aList tail}].

	"more than one arguments"
	bList := aList.
	[(bList := bList tail) isConsCell] whileTrue:
		[receiver := self send: #addLast:
						to: receiver
						with: {self argumentList: bList head}].
	bList ifNotNil:
		[^ self send: #append:
					to: receiver
					with: {self argumentList: bList}].
	^ receiver! !

!LDecompiler methodsFor: 'as yet unclassified' stamp: 'spfa 11/18/2009 18:10'!
decompile: selector in: aClass method: aMethod using: aConstructor

	((selector beginsWith: Logic prologKeyword)
		and: [selector isUnary])
	"restore the one-argument form of selector"
	ifTrue: [^ super decompile: (selector, ':') asSymbol in: aClass method: aMethod using: aConstructor].
	
	^ super decompile: selector in: aClass method: aMethod using: aConstructor

! !

!LCompiler class methodsFor: 'accessing' stamp: 'spfa 11/18/2009 18:10'!
decompilerClass

	^LDecompiler! !

!LParser methodsFor: 'prolog parser' stamp: 'spfa 11/18/2009 18:10'!
goalSeries
	"Do the work for method statements by
	 adding each goal to statements."
	" (Prolog new :? goal1(x,y), goal2(z,y)) do: [:i | ...] "
	| aList |

	aList := PrologList car: self predicate.
	[self match: #comma] whileTrue: [aList addLast: self predicate].
	^self arguments: aList
! !

!LParser methodsFor: 'error correction' stamp: 'spfa 11/18/2009 18:10'!
interactive

	^ (requestor isKindOf: REPLMorph) not
			and: [super interactive]! !

!LParser methodsFor: 'prolog parser' stamp: 'spfa 11/18/2009 18:10'!
listTerms
	"Answer a list of predicate terms."
	| term aList |
	self advance.
	(self match: #rightBracket) ifTrue: [^ nil].
	term := self term.
	aList := PrologList car: term.
	[
		(self match: #verticalBar) ifTrue: [
				aList append: self term.
				(self match: #rightBracket) ifTrue: [^ aList].
				self expected: 'Right bracket'].
		(self match: #comma)
			ifTrue: [aList addLast: self term]
			ifFalse: [
				(self match: #rightBracket) ifTrue: [^aList].
				self expected: 'Right bracket']
	] repeat! !

!LParser methodsFor: 'expression types' stamp: 'spfa 11/18/2009 18:10'!
messagePart: level repeat: repeat
	"bolot 8/5/1999 18:14"
	"This is the same as Parser>>messagePart:repeat:"
	"with only ONE addition: support for #:? for calling Prolog stuff"

	| start receiver selector args precedence words keywordStart |
	[receiver := parseNode.
	(hereType == #keyword and: [level >= 3])
		ifTrue: 
			[start := self startOfNextToken.
			selector := WriteStream on: (String new: 32).
			args := OrderedCollection new.
			words := OrderedCollection new.
			[hereType == #keyword]
				whileTrue: 
					[keywordStart := self startOfNextToken + requestorOffset.
					selector nextPutAll: self advance.
					words addLast: (keywordStart to: self endOfLastToken + requestorOffset).
					self primaryExpression ifFalse: [^self expected: 'Argument'].
					self messagePart: 2 repeat: true.
					args addLast: parseNode].
			(Symbol hasInterned: selector contents ifTrue: [ :sym | selector := sym])
				ifFalse: [ selector := self correctSelector: selector contents
										wordIntervals: words
										exprInterval: (start to: self endOfLastToken)
										ifAbort: [ ^ self fail ] ].
			precedence := 3]
		ifFalse: [((hereType == #binary or: [hereType == #verticalBar])
				and: [level >= 2])
				ifTrue: 
					[start := self startOfNextToken.
					selector := self advance asSymbol.
					self primaryExpression ifFalse: [^self expected: 'Argument'].
					self messagePart: 1 repeat: true.
					args := Array with: parseNode.
					precedence := 2]
				ifFalse: [hereType == #word
						ifTrue: 
							[start := self startOfNextToken.
							selector := self advance.
							args := #().
							words := OrderedCollection with:
									(start + requestorOffset to: self endOfLastToken + requestorOffset).
							(Symbol hasInterned: selector ifTrue: [ :sym | selector := sym])
								ifFalse: [ selector := self correctSelector: selector
													wordIntervals: words
													exprInterval: (start to: self endOfLastToken)
													ifAbort: [ ^ self fail ] ].
							precedence := 1]
						ifFalse: [(hereType == #colonQuestion)
								ifTrue: [selector := #inferSmalltalk:.
										"change typeTable for Prologging"
										typeTable at: $, asciiValue put: #comma;
											at: $!! asciiValue put: #exclamationMark.
										self advance. 
										args := Array with: (self goalSeries) .
										"restore typeTable for Smalltalking"
										typeTable at: $, asciiValue put: (TypeTable at: $, asciiValue);
											at: $!! asciiValue put: (TypeTable at: $!! asciiValue).
										precedence := 3 "TODO: pick right precedence"]
								ifFalse: [^args notNil]]]].
	parseNode := MessageNode new
				receiver: receiver
				selector: selector
				arguments: args
				precedence: precedence
				from: encoder
				sourceRange: (start to: self endOfLastToken).
	repeat]
		whileTrue: [].
	^true! !

!LParser methodsFor: 'expression types' stamp: 'spfa 12/1/2011 10:17'!
method: doit context: ctxt

    | sap blk prim messageComment methodNode |

    "need temporaries/statements in doits - to enable unbounds"
    temporaries := OrderedCollection new.
    statements := OrderedCollection new.
    "may need to change (temporarily) for doits"
    typeTable := typeTable copy.

    "check if it's a Prolog method definition"
	((doit not and: [here beginsWith: Logic prologKeyword]) 
		and: [(here last == $:) not])
     ifTrue:
            [^self prologMethod: doit context: ctxt].

    sap := self pattern: doit inContext: ctxt.
    "sap={selector, arguments, precedence}"
  "  (sap at: 2) do: [:argNode | argNode beMethodArg]."   "... or   isArg: true"
    temporaries := self temporaries asOrderedCollection.

    messageComment := currentComment.
    currentComment := nil.
    prim := doit ifTrue: [0] ifFalse: [self primitive].
    self statements: #() innerBlock: doit.

    parseNode statements: statements , parseNode statements.

    blk := parseNode.
    doit ifTrue: [blk returnLast]
        ifFalse: [blk returnSelfIfNoOther].
    hereType == #doIt ifFalse: [^self expected: 'Nothing more'].
 "   self interactive ifTrue: [self removeUnusedTemps]."

    methodNode := self newMethodNode comment: messageComment.
    ^methodNode
        selector: (sap at: 1)
        arguments: (sap at: 2)
        precedence: (sap at: 3)
        temporaries: temporaries
        block: blk
        encoder: encoder
        primitive: prim! !

!LParser methodsFor: 'expression types' stamp: 'spfa 4/15/2010 17:57'!
method: doit context: ctxt encoder: encoderToUse
	" pattern [ | temporaries ] block => MethodNode."

	encoder := encoderToUse.
	^self method: doit context: ctxt! !

!LCompiler class methodsFor: 'accessing' stamp: 'bolot 7/2/1999 17:25'!
parserClass
	"Return a parser class to use for parsing method headers."

	^LParser! !

!LParser methodsFor: 'prolog parser' stamp: 'spfa 11/18/2009 18:10'!
predCut
	cuts := true.
	^ PrologList with: #primlogicCut: 
		with: (self returnBlock: (self send: #asArray to: 'answer'))
! !

!LParser methodsFor: 'prolog parser' stamp: 'spfa 11/18/2009 18:10'!
predicate

	| result pred  |

	(hereType == #period
		or: [hereType == #comma] 
		or: [hereType == #doIt])
			ifTrue: [self expected: 'Predicate'].

	(self match: #exclamationMark) ifTrue: [^ self predCut].

	hereType == #word
		ifFalse: [self notify: 'illegal predicate name'].
	pred := here asString.

	self advance.

	(self match: #leftParenthesis)
		ifFalse: [^ Relation with: (Logic makeLogicSelector: pred)].

	(pred = 'is' or: [pred = 'consult'])
		ifFalse: [^ self structure: (Logic makeLogicSelector: pred)].

	"special cases: involve parsing smalltalk code"

	result := Relation with: (Logic makeLogicSelector: pred).
	result addLast: self term.
	(self match: #comma)
		ifFalse: [self expected: 'Term (comma?)'].
	self expression
		ifFalse: [self expected: 'Expression'].
    (self match: #rightParenthesis)
		ifFalse: [self expected: 'Right parenthesis'].
	result addLast:
		(BlockNode new
			arguments: #()
			statements: {parseNode}
			returns: false
			from: encoder).

	^result! !

!LParser methodsFor: 'prolog parser' stamp: 'spfa 11/18/2009 18:10'!
predicateSeries
	| pred |
	pred := self predicate.
    ^BlockNode new
		arguments: #()
		statements: {self send: pred head
							to: 'self'
							with: {self send: #key:value:
										to: 'Association'
										with: {
											self arguments: pred tail .
											(self match: #comma)
												ifTrue: [self predicateSeries]
												ifFalse:[self send: #value
															to: prologMethodArgument
															with: #()]}}}
		returns: false
		from: encoder! !

!LParser methodsFor: 'hack' stamp: 'spfa 11/18/2009 18:10'!
primitive

	"ugly hack for 3.8 -> 3.10 compatibility"

	^ [self pragmaPrimitives] on: Error do: [super primitive]! !

!LParser methodsFor: 'expression types' stamp: 'spfa 11/18/2009 18:10'!
prologMethod: doit context: ctxt

	| blk messageComment methodNode blockComment |

	prologMethodSelector := (self pattern: doit inContext: ctxt) first.

	typeTable at: $, asciiValue put: #comma.
	typeTable at: $!! asciiValue put: #exclamationMark.

	(prologMethodArgument := encoder bindArg: #assoc).
	" prologMethodArgument  beMethodArg. "  "... or   isArg: true"
	messageComment := currentComment.
	currentComment := nil.
	" self statements: #() innerBlock: doit. "
	parseNode := BlockNode new
				arguments: #()
				statements: (self ruleSeries, self safeGuardReturnStatements)
				returns: doit
				from: encoder.
	blockComment := currentComment.
	currentComment := nil.
	parseNode comment: blockComment.
	blk := parseNode.
	hereType == #doIt ifFalse: [^self expected: 'Nothing more'].
	methodNode := MethodNode new comment: messageComment.

	^methodNode
		selector: prologMethodSelector
		arguments: {prologMethodArgument}
		precedence: 3
		temporaries: temporaries
		block: blk
		encoder: encoder
		primitive: 0! !

!LParser methodsFor: 'error correction' stamp: 'spfa 11/18/2009 18:10'!
queryUndefined
	"a lot of Prolog variables are undefined because they are unbound"
	"let's prevent unnecessary interrogation (unless it's a Smalltalk method)"

	prologMethodSelector ifNil: [^super queryUndefined]! !

!LParser methodsFor: 'prolog parser' stamp: 'spfa 11/18/2009 18:10'!
returnBlock: anExpression
	"Answer a Block which returns anExpression."

	^BlockNode new
		arguments: #()
		statements: {
			ReturnNode new
				expr: anExpression
				encoder: encoder
				sourceRange: (1 to: 1)}
		returns: true
		from: encoder! !

!LParser methodsFor: 'prolog parser' stamp: 'spfa 11/18/2009 18:10'!
rule
	"rule:
		predicate [:- predicates] "
	| predicate block |
	cuts := false.
	(predicate := self predicate) head== prologMethodSelector
				ifFalse: [self notify: 'predicate name different'].

	block := (self match: #colonDash)
		ifTrue: [self predicateSeries]
		ifFalse: [self send: #value to: prologMethodArgument].

	(self match: #period) ifFalse: [self expected: 'Period'].
	
	^ self send: #unify:with:then:
		to: 'self'
		with: {
				self send: #key to: prologMethodArgument.
				self arguments: predicate tail.
				(hereType == #doIt and: [cuts not])
						ifTrue: [self returnBlock: block]  "optimize the stack" 
						ifFalse: block
				}! !

!LParser methodsFor: 'prolog parser' stamp: 'spfa 11/18/2009 18:10'!
ruleSeries
	"Do the work for method statements by
	 adding each rule to statements."

	[statements add: self rule. source atEnd] whileFalse.
	^statements! !

!LParser methodsFor: 'prolog parser' stamp: 'spfa 11/18/2009 18:10'!
safeGuardReturnStatements
	^ {
		self send: #ifTrue:ifFalse: 
			to: (self send: #isEmptyOrNil to: 'answer')
			with: {
					self returnBlock: (encoder encodeVariable: 'nil') .
					self returnBlock: (self send: #asArray to: 'answer')
					}
		}
	! !

!LDecompiler methodsFor: 'as yet unclassified' stamp: 'spfa 11/18/2009 18:10'!
send: selector super: superFlag numArgs: numArgs

	| sel |

	sel := selector.

	(selector isUnary
		and: [selector beginsWith: Logic prologKeyword])
	ifTrue: [
	"here selector is a Prolog predicate that will be dynamically looked up in Logic>>#doesNotUnderstand. 	we 'decompile' it by giving an abstract selector, for information only."
		sel := ('logic.', 
				(selector allButFirst: Logic prologKeyword size) 					withFirstCharacterDownshifted,
				 ':') asSymbol].

	^ super send: sel super: superFlag numArgs: numArgs
! !

!LParser methodsFor: 'prolog parser' stamp: 'spfa 11/18/2009 18:10'!
send: selector to: receiver 

	^ self send: selector to: receiver with: #()! !

!LParser methodsFor: 'prolog parser' stamp: 'spfa 11/18/2009 18:10'!
send: selector to: receiver with: arguments
	"selector is a Symbol or a Selector
	node. receiver can be
	either a String or other nodes."

	^ MessageNode new
		receiver: (receiver isString
						ifTrue: [encoder encodeVariable: receiver]
						ifFalse: [receiver])
		selector: (selector asSymbol)
		arguments: (arguments collect: [:elem |
							(elem isKindOf: ParseNode)
								ifTrue: [elem]
								ifFalse: [encoder encodeLiteral: elem]])
		precedence: 3
		from: encoder
		sourceRange: (1 to: 1)! !

!LParser methodsFor: 'prolog parser' stamp: 'spfa 11/18/2009 18:10'!
structure: aSymbol
	"Process the argument part of a structure
	 whose functor is aSymbol. Current token
	 is now the left paren. Answer a Relation
	 representing the structure."

	| result |

	result := Relation with: aSymbol.
	(self match: #rightParenthesis) ifTrue: [^ result].

	[result addLast: self term. self match: #comma] whileTrue.
	(self match: #rightParenthesis)
		ifFalse: [self expected: 'Right parenthesis'].

	^result! !

!LParser methodsFor: 'prolog parser' stamp: 'spfa 11/18/2009 18:10'!
term
	| result |
	((hereType == #rightParenthesis) or: [hereType == #comma])
		ifTrue: [self expected: 'Term'].
	(hereType == #leftBracket)
		ifTrue: [^self listTerms].
	(self match: #leftParenthesis)
		ifTrue: [^self structure: #and].
	result := here.
	(hereType == #number and: [token == #@])
		ifTrue:
			[self advance; advance.
			result := result @ here.
			self advance.
			^result].
	(hereType == #string or: [hereType == #number or: [hereType == #literal]])
		ifTrue: 
			[self advance.
			^result].
	hereType == #word
		ifFalse: [
			(self match: #leftArrow) "should be #underscore"
				ifTrue: [^self send: #new to: 'LogicDontCare']
				ifFalse: [self notify: 'illegal term']].
	self advance.
	(self match: #leftParenthesis)
		ifTrue: [^self structure: result asSymbol]
		ifFalse: [^self addTemp: result]! !

!LParser methodsFor: 'multi-character scans' stamp: 'spfa 11/18/2009 18:10'!
xColon		"process prolog :- and :? symbols"
	aheadChar = $- ifTrue:
		[self step.
		tokenType := #colonDash.
		self step.
		^ token := #colonDash].
	aheadChar = $? ifTrue:
		[self step.
		tokenType := #colonQuestion.
		self step.
		^ token := #colonQuestion].
	^super xColon! !

!LParser methodsFor: 'hack' stamp: 'spfa 12/21/2012 14:17'!
xLetter
	"Form a word or keyword."

	| type c |
	
	(SystemVersion current majorMinorVersion >= 'Squeak4.4')
		ifTrue: [^ super xLetter].
	
	buffer reset.
	[c := hereChar asciiValue.
	(type := typeTable at: c ifAbsent: [#xLetter]) == #xLetter or: [type == #xDigit]]
		whileTrue: ["open code step for speed"
			buffer nextPut: hereChar.
			hereChar := aheadChar.
			source atEnd
				ifTrue: [aheadChar := 30 asCharacter
					"doit"]
				ifFalse: [aheadChar := source next]].
	(type == #colon or: [type == #xColon and: [aheadChar ~= $=]])
		ifTrue: [buffer nextPut: self step.
			["Allow any number of embedded colons in literal symbols"
			(typeTable at: hereChar asciiValue ifAbsent: [#xLetter])
				== #xColon]
				whileTrue: [buffer nextPut: self step].
			tokenType := #keyword]
		ifFalse: [tokenType := #word].
	token := buffer contents.
	token isOctetString ifTrue: [token := token asOctetString].
! !
